from MetaConf.elementtree import ElementTree
import logging
import inspect
import sys, os

__all__ = ["Error", "getroot", "execute", "logger"]

CACHE = {}

CONFIGFILES = ["~/.metaconf/config", "/etc/metaconf"]

class Error(Exception): pass

def getlogger():
    class Formatter(logging.Formatter):
        def format(self, record):
            record.llevelname = record.levelname.lower()
            return logging.Formatter.format(self, record)
    formatter = Formatter("%(llevelname)s: %(message)s")
    handler = logging.StreamHandler(sys.stderr)
    handler.setFormatter(formatter)
    logger = logging.getLogger("metaconf")
    logger.addHandler(handler)
    return logger

logger = getlogger()

def include(element, defaultdir):
    for i in range(len(element)):
        child = element[i]
        if child.tag.lower() == "include":
            filename = child.text.strip()
            if not filename.startswith("/"):
                filename = os.path.join(defaultdir, filename)
            if not os.path.isfile(filename):
                raise Error, "included file not found: "+filename
            element[i] = ElementTree.parse(filename).getroot()
            include(element[i], defaultdir)
        include(child, defaultdir)

def getroot(opts):
    root = None
    if opts.conffile:
        if not os.path.isfile(opts.conffile):
            raise Error, "configuration file not found: "+opts.conffile
        logger.debug("parsing configuration file: "+opts.conffile)
        root = ElementTree.parse(opts.conffile).getroot()
        include(root, os.path.dirname(opts.conffile))
    else:
        for conffile in CONFIGFILES:
            conffile = os.path.expanduser(conffile)
            if os.path.isfile(conffile):
                logger.debug("parsing configuration file: "+conffile)
                root = ElementTree.parse(conffile).getroot()
                include(root, os.path.dirname(conffile))
                break
        else:
            raise Error, "no configuration file found in: " + \
                         " ".join(CONFIGFILES)
    return root

def execute(stack, element, opts):
    global CACHE
    logger.debug("starting execution of element '%s' at 0x%x" %
                 (element.tag, id(element)))
    name = element.tag.lower()
    if name not in CACHE:
        logger.debug("function not cached, loading it")
        try:
            dirmodule = __import__("MetaConf.tags."+name)
            dirmodule = getattr(dirmodule, "tags")
            module = getattr(dirmodule, name)
            func = CACHE[name] = getattr(module, "run")
        except ImportError:
            raise Error, "can't find module for '%s'" % name
        except AttributeError:
            raise Error, "module '%s' has no 'run' function" % name
    else:
        logger.debug("function is cached")
        func = CACHE[name]
    # TODO: Remove this migration code.
    if element.get("parms"):
        raise Error, "'parms' attribute has been renamed to 'parm'"
    parms = element.get("parm", "").split()
    scopeparms = {}
    if parms:
        logger.debug("element needs parameters, processing them")
        for parm in parms:
            if parm in opts.parms:
                scopeparms[parm] = opts.parms[parm]
            elif stack and parm in stack[-1].get("parms", ()):
                scopeparms[parm] = stack[-1].parms[parm]
            else:
                logger.error("%s require parameters: %s" %
                             (name, " ".join(parms)))
                if stack:
                    stack[0].error = 1
                return False
    scope = Scope()
    scope.name = name
    scope.stack = stack
    scope.element = element
    scope.opts = opts
    scope.parms = scopeparms
    kwargs = {}
    logger.debug("building arguments")
    for argname in inspect.getargs(func.func_code)[0]:
        if argname == "scope":
            kwargs[argname] = scope
        elif argname in scope:
            kwargs[argname] = scope[argname]
        else:
            raise Error, "tag '%s' expects unexistent argument '%s'" % \
                         (name, argname)
    stack.insert(0, scope)
    scope.error = 0
    logger.debug("calling tag function")
    try:
        func(**kwargs)
    except Error, e:
        logger.error(str(e))
        scope.error = 1
    stack.pop(0)
    if stack:
        stack[0].error |= scope.error
    res = (not scope.error)
    logger.debug("finishing execution of element '%s' at 0x%x (%d)" %
                 (element.tag, id(element), res))
    return res

class Scope(dict):
    def __getattr__(self, name):
        return self[name]

    def __setattr__(self, name, value):
        self[name] = value

